#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Komal Thareja (kthare10@renci.org)
"""FABRIC P4 programmable switch abstraction.

This module provides the Switch class for working with P4-programmable
network switches in FABRIC. Switches extend the Node class with specialized
functionality for programmable data planes.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from tabulate import tabulate

from fabrictestbed_extensions.fablib.constants import Constants
from fabrictestbed_extensions.fablib.interface import Interface
from fabrictestbed_extensions.fablib.node import Node

if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice

from fabrictestbed.slice_editor import Capacities
from fabrictestbed.slice_editor import Node as FimNode

log = logging.getLogger("fablib")


class Switch(Node):
    """Represents a P4-programmable network switch in FABRIC.

    Switch extends :class:`~fabrictestbed_extensions.fablib.node.Node` to model
    programmable data-plane devices with interfaces exposed as P4 ports. The
    management username defaults to the FABRIC system account for switch access.

    :ivar str username: Set to :data:`~fabrictestbed_extensions.fablib.constants.Constants.FABRIC_USER`
        for system-level access.
    """

    _show_title = "Switch"

    def __init__(
        self,
        slice: Slice,
        node: FimNode,
        validate: bool = False,
        raise_exception: bool = False,
    ):
        """
        Switch constructor, usually invoked by ``Slice.add_switch()``.

        :param slice: the fablib slice to have this switch on
        :type slice: Slice

        :param node: the FIM node that this Switch represents
        :type node: FimNode

        :param validate: Validate node can be allocated w.r.t available resources
        :type validate: bool

        :param raise_exception: Raise exception in case validation fails
        :type raise_exception: bool

        """
        super(Switch, self).__init__(
            slice=slice, node=node, validate=validate, raise_exception=raise_exception
        )
        self.username = Constants.FABRIC_USER

        # Cached interfaces
        self._interfaces_cache: Dict[str, Interface] = {}

    def _invalidate_cache(self):
        """Invalidate all cached properties including interfaces."""
        super()._invalidate_cache()
        self._interfaces_cache = {}

    def update(self, fim_node: FimNode = None):
        """
        Update the switch with new FIM data.

        :param fim_node: The new FIM node data
        :type fim_node: FimNode
        """
        if fim_node:
            self.fim_node = fim_node
            self._invalidate_cache()
            self.get_interfaces(refresh=True)
            self._fim_dirty = False

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the
        node.

        Intended for printing node information.

        :return: Tabulated string of node information
        :rtype: String
        """
        table = [
            ["ID", self.get_reservation_id()],
            ["Name", self.get_name()],
            ["Site", self.get_site()],
            ["Management IP", self.get_management_ip()],
            ["Reservation State", self.get_reservation_state()],
            ["Error Message", self.get_error_message()],
            ["SSH Command", self.get_ssh_command()],
        ]

        return tabulate(table)  # , headers=["Property", "Value"])

    @staticmethod
    def get_switch(slice: Slice, node: FimNode) -> Switch:
        """
        Factory method to create a Switch from a FIM node.

        :param slice: the slice this switch belongs to
        :type slice: Slice
        :param node: the FIM node
        :type node: FimNode
        :return: a Switch instance
        :rtype: Switch
        """
        switch = Switch(slice=slice, node=node)
        switch.get_interfaces()
        return switch

    @staticmethod
    def new_switch(
        slice: Slice = None,
        name: str = None,
        site: str = None,
        avoid: List[str] = None,
        validate: bool = False,
        raise_exception: bool = False,
    ) -> Switch:
        """
        Creates a new FABRIC switch on the slice.

        Not intended for API use. Use slice.add_switch() instead.

        :param slice: the fablib slice to build the new switch on
        :type slice: Slice
        :param name: the name of the new switch
        :type name: str
        :param site: the name of the site to build the switch on
        :type site: str
        :param avoid: a list of site names to avoid
        :type avoid: List[str]
        :param validate: Validate switch can be allocated w.r.t available resources
        :type validate: bool
        :param raise_exception: Raise exception if validation fails
        :type raise_exception: bool
        :return: a new Switch
        :rtype: Switch
        """
        if avoid is None:
            avoid = []

        if site is None:
            [site] = slice.get_fablib_manager().get_random_sites(avoid=avoid)

        log.info(f"Adding switch: {name}, slice: {slice.get_name()}, site: {site}")

        switch = Switch(
            slice,
            slice.topology.add_switch(name=name, site=site),
            validate=validate,
            raise_exception=raise_exception,
        )
        # Set capacities
        cap = Capacities(unit=1)
        switch.get_fim().set_properties(capacities=cap)
        switch.init_fablib_data()

        return switch

    @staticmethod
    def get_pretty_name_dict():
        """
        Get a mapping of field names to human-readable labels for display.

        Returns a dictionary that maps internal field names to user-friendly
        display names used when rendering tables and formatted output.

        :return: Dictionary mapping field names to pretty names
        :rtype: dict
        """
        return {
            "id": "ID",
            "name": "Name",
            "site": "Site",
            "username": "Username",
            "management_ip": "Management IP",
            "state": "State",
            "error": "Error",
            "ssh_command": "SSH Command",
            "public_ssh_key_file": "Public SSH Key File",
            "private_ssh_key_file": "Private SSH Key File",
        }

    def toDict(self, skip: list = None):
        """
        Returns the node attributes as a dictionary.

        Results are cached. Cache is invalidated when ``_invalidate_cache()``
        is called.

        :param skip: list of keys to exclude
        :type skip: list
        :return: switch attributes as dictionary
        :rtype: dict
        """
        if skip is None:
            skip = []

        if self._cached_dict is None:
            d = {}
            d["id"] = str(self.get_reservation_id())
            d["name"] = str(self.get_name())
            d["site"] = str(self.get_site())
            d["username"] = str(self.get_username())
            d["management_ip"] = (
                str(self.get_management_ip()).strip()
                if str(self.get_reservation_state()) == "Active"
                and self.get_management_ip()
                else ""
            )
            d["state"] = str(self.get_reservation_state())
            d["error"] = str(self.get_error_message())
            if str(self.get_reservation_state()) == "Active":
                d["ssh_command"] = str(self.get_ssh_command())
            else:
                d["ssh_command"] = ""
            d["public_ssh_key_file"] = str(self.get_public_key_file())
            d["private_ssh_key_file"] = str(self.get_private_key_file())
            self._cached_dict = d

        if not skip:
            return dict(self._cached_dict)
        return {k: v for k, v in self._cached_dict.items() if k not in skip}

    def generate_template_context(self, skip: list = None):
        """
        Generate the base template context for this switch.

        Creates a dictionary context suitable for Jinja2 template rendering,
        excluding the SSH command and setting an empty components list.

        :param skip: list of keys to exclude
        :type skip: list
        :return: Template context dictionary with switch attributes
        :rtype: dict
        """
        if skip is None:
            skip = ["ssh_command"]
        else:
            skip = list(skip)
            if "ssh_command" not in skip:
                skip.append("ssh_command")
        context = self.toDict(skip=skip)
        context["components"] = []
        return context

    def get_fim(self) -> FimNode:
        """
        Not recommended for most users.

        Gets the node's FABRIC Information Model (fim) object. This method
        is used to access data at a lower level than FABlib.

        :return: the FABRIC model node
        :rtype: FIMNode
        """
        return self.fim_node

    def __set_capacities(self, unit: int = 1):
        """
        Sets the capacities of the FABRIC node.
        """
        cap = Capacities(unit=unit)
        self.get_fim().set_properties(capacities=cap)

    def delete(self):
        """
        Remove the switch from the slice. All components and interfaces associated with
        the Node are removed from the Slice.
        """
        self.get_slice().get_fim_topology().remove_switch(name=self.get_name())

    def get_interfaces(
        self, include_subs: bool = True, refresh: bool = False, output: str = "list"
    ) -> Union[Dict[str, Interface], List[Interface]]:
        """
        Gets a list of the interfaces associated with the FABRIC node.

        Results are cached. Use refresh=True to force reload.

        :param include_subs: Flag indicating if sub interfaces should be included
        :type include_subs: bool

        :param refresh: Refresh the interface object with latest Fim info
        :type refresh: bool

        :param output: return type - 'list' or 'dict'
        :type output: str

        :return: interfaces on the node
        :rtype: Union[Dict[str, Interface], List[Interface]]
        """
        if self._interfaces_cache and not refresh and not self._fim_dirty:
            if output == "dict":
                return self._interfaces_cache
            return [
                self._interfaces_cache[key]
                for key in sorted(self._interfaces_cache.keys())
            ]

        self._interfaces_cache = {}

        try:
            if self.fim_node and hasattr(self.fim_node, "interfaces"):
                for name, fim_iface in self.fim_node.interfaces.items():
                    self._interfaces_cache[name] = Interface(
                        node=self, fim_interface=fim_iface, model="NIC_P4"
                    )
        except Exception as e:
            log.debug(f"Error getting interfaces: {e}")

        # Keep self.interfaces in sync for backward compatibility
        self.interfaces = dict(self._interfaces_cache)

        if output == "dict":
            return self._interfaces_cache

        return [
            self._interfaces_cache[key] for key in sorted(self._interfaces_cache.keys())
        ]

    def get_interface(
        self, name: str = None, refresh: bool = False
    ) -> Optional[Interface]:
        """
        Gets a specific interface by name.

        :param name: the interface name
        :type name: str
        :param refresh: force refresh from FIM
        :type refresh: bool
        :return: the interface
        :rtype: Interface
        """
        if not self._interfaces_cache or refresh or self._fim_dirty:
            self.get_interfaces(refresh=refresh, output="dict")

        return self._interfaces_cache.get(name)

    @staticmethod
    def get_node(slice: Slice = None, node=None):
        """
        Returns a new fablib node using existing FABRIC resources.

        :note: Not intended for API call.

        :param slice: the fablib slice storing the existing node
        :type slice: Slice

        :param node: the FIM node stored in this fablib node
        :type node: Node

        :return: a new fablib node storing resources
        :rtype: Node
        """
        ret_val = Switch(slice, node)
        ret_val.get_interfaces()
        return ret_val



from __future__ import annotations

import json
from abc import abstractmethod
from typing import List, Optional

import jinja2
from fabrictestbed.slice_editor import UserData

from fabrictestbed_extensions.utils.utils import Utils


class TemplateMixin:
    """Mixin providing Jinja2 template rendering for FABRIC resource classes.

    Subclasses may override:

    * ``_default_skip`` – list of field names excluded from the template
      context by default (e.g. ``["ssh_command"]`` for nodes).
    * ``generate_template_context()`` – builds the dict that represents
      *this* object inside the template namespace.
    * ``_configure_template_environment(environment)`` – hook for
      customising the :class:`jinja2.Environment` before rendering.
    """

    _default_skip: Optional[List[str]] = None
    _show_title: str = ""

    def __init__(self, **kwargs):
        # V2 specific: dirty flag for caching
        self._fim_dirty: bool = True

        self._cached_reservation_id: Optional[str] = None
        self._cached_reservation_state: Optional[str] = None
        self._cached_error_message: Optional[str] = None
        self._cached_name: Optional[str] = None
        self._cached_dict: Optional[dict] = None

    def _invalidate_cache(self):
        """
        Invalidate all cached properties.

        Called when the FIM node is updated.
        """
        self._fim_dirty = True
        self._cached_reservation_id = None
        self._cached_reservation_state = None
        self._cached_error_message = None
        self._cached_name = None
        self._cached_dict = None

    @abstractmethod
    def get_fim(self):
        """
        Returns fim of the template.
        """

    @abstractmethod
    def toDict(self, skip: Optional[List[str]]):
        """
        Returns the attributes as a dictionary

        :return: attributes as dictionary
        :rtype: dict
        """

    def generate_template_context(self, skip: Optional[List[str]] = None):
        """Return a dict representing this object for template rendering.

        The default implementation delegates to ``self.toDict()``.
        """
        return self.toDict(skip=skip)

    def _configure_template_environment(self, environment: jinja2.Environment):
        """Hook to customise the Jinja2 environment before rendering."""
        pass

    @abstractmethod
    def get_slice(self):
        """Return a :class:`Slice` object representing the current slice."""

    def get_template_context(self, skip=None):
        """Get the full Jinja2 template context from the parent slice.

        :param skip: Field names to exclude. Falls back to
            ``_default_skip`` when *None*.
        :type skip: list[str] | None
        :return: Template context dictionary for Jinja2 rendering.
        :rtype: dict
        """
        effective_skip = skip if skip is not None else self._default_skip
        if effective_skip:
            return self.get_slice().get_template_context(self, skip=effective_skip)
        return self.get_slice().get_template_context(self)

    def render_template(self, input_string, skip=None):
        """Render a Jinja2 template string using this object's context.

        :param input_string: Jinja2 template string to render.
        :type input_string: str
        :param skip: Field names to exclude from the context.
        :type skip: list[str] | None
        :return: Rendered template output string.
        :rtype: str
        """
        environment = jinja2.Environment()
        self._configure_template_environment(environment)
        template = environment.from_string(input_string)
        return template.render(self.get_template_context(skip=skip))

    def get_error_message(self) -> str:
        """
        Gets the error messages

        :return: network service types
        :rtype: String
        """
        if self._cached_error_message is None:
            try:
                self._cached_error_message = str(
                    self.get_fim().get_property(pname="reservation_info").error_message
                )
            except Exception:
                self._cached_error_message = None
        return self._cached_error_message

    def get_reservation_id(self) -> Optional[str]:
        """
        Gets the reservation ID of the node.

        Results are cached for performance.

        :return: reservation ID
        :rtype: String
        """
        if self._cached_reservation_id is None:
            try:
                self._cached_reservation_id = str(
                    self.get_fim().get_property(pname="reservation_info").reservation_id
                )
            except Exception:
                self._cached_reservation_id = None
        return self._cached_reservation_id

    def get_reservation_state(self) -> Optional[str]:
        """
        Gets the reservation state on the FABRIC node.

        :return: the reservation state on the node
        :rtype: String
        """
        if self._cached_reservation_state is None:
            try:
                self._cached_reservation_state = str(
                    self.get_fim()
                    .get_property(pname="reservation_info")
                    .reservation_state
                )
            except Exception:
                self._cached_reservation_state = None
        return self._cached_reservation_state

    def get_name(self) -> str:
        """
        Gets the name of this network service.

        Results are cached for performance.

        :return: the name of this network service
        :rtype: String
        """
        if self._cached_name is None:
            try:
                if self.get_fim():
                    self._cached_name = self.get_fim().name
                else:
                    self._cached_name = None
            except Exception:
                self._cached_name = None
        return self._cached_name

    def toJson(self):
        """
        Returns the attributes as a JSON string.

        :return: attributes as JSON string
        :rtype: str
        """
        return json.dumps(self.toDict(), indent=4)

    def get_fablib_manager(self):
        """
        Get a reference to :py:class:`.FablibManager`.
        """
        s = self.get_slice()
        return s.get_fablib_manager() if s is not None else None

    @staticmethod
    def get_pretty_name_dict():
        """
        Return a mapping of internal field names to display names.

        Subclasses should override this to provide their own mappings.

        :return: pretty name mapping
        :rtype: dict
        """
        return {}

    def show(
        self, fields=None, output=None, quiet=False, colors=False, pretty_names=True
    ):
        """
        Show a table containing the current attributes.

        There are several output options: ``"text"``, ``"pandas"``,
        and ``"json"`` that determine the format of the output that is
        returned and (optionally) displayed/printed.

        :param output: output format
        :type output: str
        :param fields: list of fields to show
        :type fields: List[str]
        :param quiet: True to specify printing/display
        :type quiet: bool
        :param colors: True to specify state colors for pandas output
        :type colors: bool
        :param pretty_names: Display pretty names
        :type pretty_names: bool
        :return: table in format specified by output parameter
        :rtype: Object
        """
        data = self.toDict()

        if pretty_names:
            pretty_names_dict = self.get_pretty_name_dict()
        else:
            pretty_names_dict = {}

        table = Utils.show_table(
            data,
            fields=fields,
            title=self._show_title,
            output=output,
            quiet=quiet,
            pretty_names_dict=pretty_names_dict,
        )

        return table

    def set_user_data(self, user_data: dict):
        """
        Set user data.

        :param user_data: a ``dict``.
        :type user_data: dict
        """
        self.get_fim().set_property(
            pname="user_data", pval=UserData(json.dumps(user_data))
        )

    def get_user_data(self) -> dict:
        """
        Get user data.

        :return: user data dictionary
        :rtype: dict
        """
        try:
            return json.loads(str(self.get_fim().get_property(pname="user_data")))
        except Exception:
            return {}

    def get_fablib_data(self) -> dict:
        """
        Get fablib data. Usually used internally.

        :return: fablib data dictionary
        :rtype: dict
        """
        try:
            return self.get_user_data()["fablib_data"]
        except Exception:
            return {}

    def set_fablib_data(self, fablib_data: dict):
        """
        Set fablib data. Usually used internally.

        :param fablib_data: fablib data dictionary
        :type fablib_data: dict
        """
        user_data = self.get_user_data()
        user_data["fablib_data"] = fablib_data
        self.set_user_data(user_data)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

"""
This module contains methods to work with FABRIC `facility ports`_.

.. _`facility ports`: https://learn.fabric-testbed.net/knowledge-base/glossary/#facility_port
"""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from fabrictestbed.slice_editor import Capacities, Labels
from tabulate import tabulate

from fabrictestbed_extensions.fablib.interface import Interface
from fabrictestbed_extensions.fablib.template_mixin import TemplateMixin

if TYPE_CHECKING:
    from fim.user.node import Node as FimNode

    from fabrictestbed_extensions.fablib.slice import Slice

log = logging.getLogger("fablib")


class FacilityPort(TemplateMixin):
    """
    A class for working with FABRIC facility ports.
    """

    _show_title = "Facility Port"

    fim_object = None
    slice = None

    def __init__(self, slice: Slice, fim_object: FimNode):
        """
        :param slice: the fablib slice to have this node on
        :type slice: Slice

        :param fim_object:
        :type fim_object: FimNode
        """
        super().__init__()
        self.fim_object: FimNode = fim_object
        self.slice: Slice = slice
        self.interfaces: Dict[str, Interface] = {}

        self._cached_site: Optional[str] = None

        # V2 specific: cached interfaces
        self._interfaces_cache: Dict[str, Interface] = {}

    def _invalidate_cache(self):
        """Invalidate all cached properties."""
        super(FacilityPort, self)._invalidate_cache()

        self._cached_site = None
        self._interfaces_cache = {}

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the
        node.  Intended for printing node information.

        :return: Tabulated string of node information
        :rtype: String
        """
        table = [["name", self.get_name()]]

        return tabulate(table)

    @staticmethod
    def get_pretty_name_dict():
        """
        Return a mapping used when rendering table headers.
        """
        return {
            "name": "Name",
        }

    def toDict(self, skip: list = None):
        """
        Return a Python `dict` representation of the facility port.

        Results are cached. Cache is invalidated when ``_invalidate_cache()``
        is called.

        :param skip: list of keys to exclude
        :type skip: list
        """
        if skip is None:
            skip = []

        if self._cached_dict is None:
            d = {}
            d["name"] = str(self.get_name())
            self._cached_dict = d

        if not skip:
            return dict(self._cached_dict)
        return {k: v for k, v in self._cached_dict.items() if k not in skip}

    def get_fim(self):
        """
        Gets the Facility Ports's FABRIC Information Model (fim) object.

        This method is used to access data at a lower level than
        FABlib.
        """
        return self.fim_object

    def get_model(self) -> str:
        """
        Get fablib model name for the facility port.
        """
        return "Facility_Port"

    def get_site(self) -> str:
        """
        Gets the site where the facility port is located.

        Results are cached for performance.

        :return: the site name
        :rtype: str
        """
        if self._cached_site is None:
            try:
                if self.fim_object and self.fim_object.site:
                    self._cached_site = self.fim_object.site
            except Exception:
                pass
        return self._cached_site if self._cached_site else ""

    @staticmethod
    def new_facility_port(
        slice: Slice = None,
        name: str = None,
        site: str = None,
        vlan: Union[List, str] = None,
        bandwidth: int = None,
        mtu: int = None,
        labels: Labels = None,
        peer_labels: Labels = None,
    ) -> FacilityPort:
        """
        Create a new facility port in the given slice.

        You might want to :py:meth:`Slice.add_facility_port()`, in
        most cases.

        :param slice: The slice in which the facility port will be created.
        :param name: The name of the facility port.
        :param site: The site where the facility port will be located.
        :param vlan: A list or single string representing the VLANs for the facility port.
        :param bandwidth: The bandwidth capacity for the facility port, default is 10.
        :param mtu: MTU size
        :param labels: Labels associated with the facility port.
        :param peer_labels: Peer labels associated with the facility port.
        :return: A FacilityPort object representing the created facility port.
        """
        capacities = Capacities()
        if bandwidth:
            capacities.bw = int(bandwidth)
        if mtu:
            capacities.mtu = mtu

        interfaces = None

        if vlan:
            index = 1
            interfaces = []
            if isinstance(vlan, str):
                vlan = [vlan]

            for v in vlan:
                iface_tuple = (
                    f"iface-{index}",
                    Labels(vlan=v),
                    capacities,
                )
                interfaces.append(iface_tuple)

        fim_facility_port = slice.get_fim_topology().add_facility(
            name=name,
            site=site,
            capacities=capacities,
            labels=labels,
            peer_labels=peer_labels,
            interfaces=interfaces,
        )
        return FacilityPort(slice, fim_facility_port)

    @staticmethod
    def get_facility_port(slice: Slice = None, facility_port: FimNode = None):
        """
        Create a FacilityPort object from an existing FIM facility port.

        Factory method that wraps a FIM (FABRIC Information Model) facility
        port object in a FABlib FacilityPort wrapper.

        :param slice: The slice containing the facility port.
        :type slice: Slice

        :param facility_port: FIM facility port object to wrap.
        :type facility_port: FimNode

        :return: FacilityPort object wrapping the FIM facility port.
        :rtype: FacilityPort
        """
        return FacilityPort(slice=slice, fim_object=facility_port)

    def get_slice(self) -> Slice:
        """
        Gets the fablib slice associated with this node.

        :return: the fablib slice on this node
        :rtype: Slice
        """
        return self.slice

    def get_interfaces(
        self, refresh: bool = False, output: str = "list"
    ) -> Union[Dict[str, Interface], List[Interface]]:
        """
        Gets interfaces associated with this facility port.

        Results are cached. Use refresh=True to force reload.

        :param refresh: force refresh from FIM
        :type refresh: bool
        :param output: return type - 'list' or 'dict'
        :type output: str
        :return: interfaces
        :rtype: Union[Dict[str, Interface], List[Interface]]
        """
        from fabrictestbed_extensions.fablib.interface import Interface

        if self._interfaces_cache and not refresh and not self._fim_dirty:
            if output == "dict":
                return self._interfaces_cache
            return list(self._interfaces_cache.values())

        self._interfaces_cache = {}

        try:
            if self.fim_object and hasattr(self.fim_object, "interfaces"):
                for iface in self.fim_object.interfaces.values():
                    interface = Interface(fim_interface=iface, node=self)
                    self._interfaces_cache[interface.get_name()] = interface
        except Exception as e:
            log.debug(f"Error getting interfaces: {e}")

        if output == "dict":
            return self._interfaces_cache
        return list(self._interfaces_cache.values())

    def get_interface(
        self, name: str = None, refresh: bool = False
    ) -> Optional[Interface]:
        """
        Gets a specific interface by name.

        :param name: the interface name
        :type name: str
        :param refresh: force refresh from FIM
        :type refresh: bool
        :rtype: Interface
        """
        # Ensure cache is populated
        interfaces = self.get_interfaces(refresh=refresh, output="dict")
        return interfaces.get(name)

    def update(self, fim_node: FimNode = None):
        """
        Update the facility port with new FIM data.

        :param fim_node: The new FIM node data
        :type fim_node: FimNode
        """
        if fim_node:
            self.fim_object = fim_node
            self._invalidate_cache()
            try:
                self.get_interfaces(refresh=True)
            except Exception as e:
                log.debug(
                    f"FacilityPort {self.get_name()}: error refreshing "
                    f"caches during update: {e}"
                )
            self._fim_dirty = False

    def delete(self):
        """
        Remove the facility from the slice. All interfaces associated with
        the Facility Port are removed from the Slice.
        """
        for iface in self.get_interfaces():
            iface.delete()

        self.get_slice().get_fim_topology().remove_facility(name=self.get_name())
